"""A media playlist library for Python

This library can be used to create complex, nested media playlists.

Modules
-------
playlist:
    defines the Playlist class, the main feature of the package
config:
    provides global configuration to the package
sequence:
    low level iterator classes
sample:
    low level random sampling routines
"""

__version__ = "1.1.0"

from . import config
from . import sample
from . import sequence
from . import playlist

Playlist = playlist.Playlist
ensure_state = playlist.ensure_state



"""Implementation of sequence subpackage"""

__all__ = [
    "AbstractSequence",
    "TrivialSequence",
    "FiniteSequence",
    "OrderedLoopingSequence",
    "ShuffledLoopingSequence",
]

from collections.abc import Iterable, Iterator
from abc import abstractmethod
import random

from nextsong.sample import sublist, weighted_choice


DEFAULT_WEIGHT = 1.0
DEFAULT_RECENT_PORTION = 0.5


class AbstractSequence(Iterable):
    """Abstract class for sequences

    Requires Iterable methods and a 'weight' property, as well as a
    can_produce method used to avoid infinite busy loops on looping
    sequences.
    """

    @property
    @abstractmethod
    def weight(self):
        """Weight relative to other AbstractWeightedIterable objects

        For example, weight could be the relative likelyhood of being
        included in a random sample.
        """

    @abstractmethod
    def can_produce(self):
        """Check whether the sequence has a chance of produce an item

        Used to prevent infinite loops
        """


class TrivialSequence(AbstractSequence):
    """An iterable that yields one item one time

    This sequence is used as a component in other sequences in order to
    simplify their logic
    """

    class _TrivialIterator(Iterator):
        # pylint: disable=too-few-public-methods
        # Reason: this is implemented as an iterator class instead of
        # generator so that it can be pickled.
        def __init__(self, item):
            self.item = item
            self.consumed = False

        def __next__(self):
            if self.consumed:
                raise StopIteration
            self.consumed = True
            return self.item

    def __init__(self, item):
        self.__item = item

    def can_produce(self):
        return True

    @property
    def weight(self):
        return DEFAULT_WEIGHT

    def __iter__(self):
        return self._TrivialIterator(self.__item)


class FiniteSequence(AbstractSequence):
    """A sequence that terminates after a pass through its items"""

    class _FiniteIterator(Iterator):
        # pylint: disable=too-few-public-methods
        # Reason: this is implemented as an iterator class instead of
        # generator so that it can be pickled.
        def __init__(self, items):
            self.stack = [iter(x) for x in reversed(items)]

        def __next__(self):
            while self.stack:
                try:
                    return next(self.stack[-1])
                except StopIteration:
                    self.stack.pop()
            raise StopIteration

    @staticmethod
    def __determine_count(item_count, portion, count):
        if portion is not None and count is not None:
            raise ValueError("portion and count are mutually exclusive")

        if count is None and portion is None:
            portion = 1

        if portion is not None:
            if isinstance(portion, (int, float)):
                portion = (portion, portion)
            if isinstance(portion, (tuple, list)):
                if len(portion) != 2 or not all(
                    isinstance(x, (int, float)) for x in portion
                ):
                    raise ValueError("portion should contain two numbers")
            else:
                raise ValueError("portion should be a number or pair of numbers")
            count = tuple(int(round(x * item_count)) for x in portion)

        if isinstance(count, int):
            count = (count, count)
        if isinstance(count, (tuple, list)):
            if len(count) != 2 or not all(isinstance(x, int) for x in count):
                raise ValueError("count should contain two ints")
        else:
            raise ValueError("count should be an int or pair of ints")
        return tuple(min(item_count, max(0, x)) for x in count)

    def __init__(self, *items, weight=None, portion=None, count=None, shuffle=False):
        items = [
            x if isinstance(x, AbstractSequence) else TrivialSequence(x) for x in items
        ]
        self.__items = items
        self.__shuffle = shuffle

        self.__count = self.__determine_count(len(items), portion, count)
        if weight is None:
            weight = DEFAULT_WEIGHT

        self.__weight = weight

    @property
    def weight(self):
        return self.__weight

    def can_produce(self):
        if not any(x.can_produce() for x in self.__items):
            return False
        if not min(*self.__count):
            return False
        return True

    def __iter__(self):
        count = random.randint(*self.__count)
        weights = [item.weight for item in self.__items]
        choices = sublist(self.__items, count, weights=weights)
        if self.__shuffle:
            random.shuffle(choices)
        return self._FiniteIterator(choices)


class OrderedLoopingSequence(AbstractSequence):
    """A sequence that repeatedly loops through its items in order"""

    class _OrderedLoopingIterator(Iterator):
        # pylint: disable=too-few-public-methods
        # Reason: this is implemented as an iterator class instead of
        # generator so that it can be pickled
        def __init__(self, sequence):
            self.sequence = sequence
            self.iterator = None

        def __next__(self):
            if not self.sequence.can_produce():
                raise StopIteration

            while True:
                if self.iterator is None:
                    self.iterator = iter(self.sequence)
                try:
                    return next(self.iterator)
                except StopIteration:
                    self.iterator = None

    def __init__(self, *items, portion=None, count=None, weight=None):
        self.__sequence = FiniteSequence(
            *items, portion=portion, count=count, shuffle=False, weight=weight
        )

    def can_produce(self):
        return self.__sequence.can_produce()

    @property
    def weight(self):
        return self.__sequence.weight

    def __iter__(self):
        return self._OrderedLoopingIterator(self.__sequence)


class ShuffledLoopingSequence(AbstractSequence):
    """A sequence that repeatedly samples randomly from its items"""

    class _ShuffledLoopingIterator(Iterator):
        # pylint: disable=too-few-public-methods
        # This is implemented as an iterator class instead of generator
        # so that it can be pickled
        def __init__(self, items, recent_size):
            self.fresh_items = list(items)
            self.recent_items = []
            self.current_iter = None
            self.recent_size = recent_size
            self.item_count = len(items)

        def __next__(self):
            all_items = self.fresh_items + self.recent_items
            while True:
                if self.current_iter is None:
                    if not any(x.can_produce() for x in all_items):
                        raise StopIteration
                    if self.fresh_items:
                        i = weighted_choice([x.weight for x in self.fresh_items])
                        choice = self.fresh_items.pop(i)
                    elif self.recent_items:
                        choice = self.recent_items.pop(0)
                    else:
                        raise RuntimeError("Unexpected logic error")
                    self.recent_items.append(choice)
                    if len(self.recent_items) > self.recent_size:
                        self.fresh_items.append(self.recent_items.pop(0))
                    self.current_iter = iter(choice)

                try:
                    return next(self.current_iter)
                except StopIteration:
                    self.current_iter = None

    def __init__(self, *items, recent_portion=None, weight=None):
        items = [
            x if isinstance(x, AbstractSequence) else TrivialSequence(x) for x in items
        ]
        if recent_portion is None:
            recent_portion = DEFAULT_RECENT_PORTION
        self.__items = items
        self.__recent_size = int(round(min(1.0, max(0.0, recent_portion)) * len(items)))

        if weight is None:
            weight = DEFAULT_WEIGHT
        self.__weight = weight

    @property
    def weight(self):
        return self.__weight

    def can_produce(self):
        if not any(x.can_produce() for x in self.__items):
            return False
        return True

    def __iter__(self):
        return self._ShuffledLoopingIterator(self.__items, self.__recent_size)



"""Implementation of sample subpackage"""

__all__ = ["weighted_choice", "sublist"]

import random


def weighted_choice(weights):
    """Picks an index at random based on weights

    This function picks a random index into the given list of weights,
    with larger weights being proportionally more likely to be chosen.
    For example, if the list of weights is [1, 2], the function has a
    1/3 chance of returning 0 and a 2/3 chance of returning 1.

    Arguments
    ---------
    weights:
        A list of nonnegative numbers.
    """
    if any(w < 0 for w in weights):
        raise ValueError("weights must be nonnegative")
    if all(w == 0 for w in weights):
        weights = [1] * len(weights)
    total_weight = sum(weights)
    normed_weights = [weight / total_weight for weight in weights]
    running_total = 0
    cumulative_weights = []
    for weight in normed_weights:
        running_total += weight
        cumulative_weights.append(running_total)
    # Ensure the final total is 1 regardless of precision issues
    cumulative_weights[-1] = 1
    threshold = random.random()
    # Choose the first item whose cumulative normalized weight exceeds
    # the random number
    for i, weight in enumerate(cumulative_weights):
        if weight > threshold:
            return i
    raise RuntimeError("Unexpected logic error")


def sublist(items, count, weights=None):
    """Extracts a sublist containing count entries from items

    Parameters
    ----------
    items: List[Any]
        The list of items to be sampled from.
    count: int
        The number of entries from items to be returned.
    weights: List[float]
        A list with equal length to items. Indicates the relative
        likelyhood of the item being used in the sample. If not given,
        all items have equal weight.

    """
    items = list(items)
    if not items:
        return []

    if not weights:
        weights = [1] * len(items)
    weights = list(weights)

    if len(items) != len(weights):
        raise ValueError("Length mismatch between items and weights")

    candidates = [
        (i, item, weight) for i, (item, weight) in enumerate(zip(items, weights))
    ]
    choices = []

    while candidates and len(choices) < count:
        weights = [weight for _, _, weight in candidates]
        i = weighted_choice(weights)
        choices.append(candidates.pop(i))

    return [item for _, item, _ in sorted(choices)]



"""Implementation of playlist subpackage"""

__all__ = ["Playlist", "ensure_state"]

from pathlib import Path
import pickle
import warnings

from lxml import etree

import nextsong.sequence as seq
from nextsong.config import get as get_config


class Playlist:
    """A class used to describe and iterate over a media playlist

    This is the nextsong library's primary class. The general workflow
    is to create a Playlist instance then iterate over it using the
    builtin iter and next functions. A Playlist may also be saved and
    loaded from an XML file using the save_xml method and load_xml
    static method.
    """

    class PlaylistState:
        """Represents the state of iteration through a Playlist

        This class is an iterator on a Playlist and can be pickled
        to a file.

        A PlaylistState can be used as the subject of a 'with'
        statement to ensure the pickle file is written at the end of
        the 'with' block. If the state was originally loaded from a
        file, the same file will be written to. Otherwise, the
        "state_path" config value will be used.
        """

        def __init__(self, iterator, from_path=None):
            """Internal constructor for PlaylistState

            Users should instead call Playlist.__iter__ or
            Playlist.load_state to create a PlaylistState.
            """
            self.__iterator = iterator
            self.__from_path = from_path

        def __next__(self):
            return next(self.__iterator)

        def save(self, filepath=None):
            """Pickles a PlaylistState instance to a file

            Arguments
            ---------
            filepath: str or None
                The desired path to pickle to. If None, uses the value
                of the "state_path" config.
            """
            if filepath is None:
                filepath = get_config("state_path")
            with open(filepath, "wb") as file:
                return pickle.dump(self, file)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.save(self.__from_path)

        @staticmethod
        def load(filepath=None, *, handle_not_found=False):
            """Loads a pickled PlaylistState instance from a file

            Arguments
            ---------
            filepath: str or None
                The path to the pickle file. If None, uses the value of
                the "state_path" config.
            handle_not_found: bool
                Determines what should happen if the given filepath
                doesn't exist. If True, a value of None is returned.
                If False, the FileNotFoundError is allowed to
                propagate.
            """
            if filepath is None:
                filepath = get_config("state_path")
            try:
                with open(filepath, "rb") as file:
                    return pickle.load(file)
            except FileNotFoundError:
                if handle_not_found:
                    return None
                raise

    def __init__(
        self,
        *children,
        shuffle=None,
        loop=None,
        portion=None,
        count=None,
        recent_portion=None,
        weight=None,
    ):
        """Create a new Playlist instance

        Positional arguments are items in the playlist. Each item should
        be either another Playlist instance or a string. A string should
        be an absolute path or a path relative to the globally
        configured 'media_root'. A string may also be a glob pattern. If
        the literal file cannot be found, it will be expanded to a list
        of files under the pathlib glob expansion rules.

        Keyword arguments
        -----------------
        shuffle:
            If True, the Playlist will iterate through its children in a
            random order.
        loop:
            If True, the Playlist will loop forever through its
            children.
        portion:
            A number or pair of numbers between 0 and 1. Specifies the
            portion of items in the playlist to be used in a pass of the
            Playlist. For example a portion of 0.2 means only 20% of
            items in the Playlist will be used. If a pair of numbers is
            given, a uniformly random value between the pair of numbers
            will be used. This argument is mutually exclusive with
            count.
        count:
            An integer or pair of integers. Specifies the number of
            items in the playlist to be used in a pass of the Playlist.
            For example a count of 2 means only 2 of items in the
            Playlist will be used. If a pair of numbers is given, a
            uniformly random value between the pair of numbers
            (inclusive) will be used. This argument is mutually
            exclusive with portion.
        recent_portion:
            A number between 0 and 1. This argument can only be used if
            shuffle and loop are both True. By default, a shuffled
            looping Playlist doesn't select items truly independently.
            Items that were recently selected are marked as recent, and
            only non-recent items are candidates for selection. This
            argument specifies the maximum portion of the Playlist that
            can be recent. For example, if recent_portion is 0.2 and the
            Playlist has 10 items, the two most recently selected items
            will not be randomly selected.
        weight:
            A non-negative number indicating the relative likelyhood of
            this Playlist being chosen over its siblings during a random
            sampling. A Playlist with a weight of zero is disabled and
            will never be selected.
        """
        self.__validate_children(children)
        self.__children = children

        options = {
            "shuffle": shuffle,
            "portion": portion,
            "count": count,
            "recent_portion": recent_portion,
            "weight": weight,
            "loop": loop,
        }
        self.__validate_options(options)
        self.__options = options

    @staticmethod
    def __validate_children(children):
        for child in children:
            if isinstance(child, Playlist):
                pass
            elif isinstance(child, str):
                pass
            else:
                raise ValueError(f"child {repr(child)} of unknown type")

    @staticmethod
    def __validate_options(options):
        if options["loop"]:
            if options["shuffle"]:
                if options["count"] is not None:
                    raise ValueError("count requires loop=False or shuffle=False")
                if options["portion"] is not None:
                    raise ValueError("portion requires loop=False or shuffle=False")
            else:
                if options["recent_portion"] is not None:
                    raise ValueError("recent_portion requires shuffle=True")
        else:
            if options["recent_portion"] is not None:
                if options["shuffle"]:
                    raise ValueError("recent_portion requires loop=True")
                raise ValueError("recent_portion requires loop=True and shuffle=True")

    @property
    def children(self):
        """The items iterated over by the Playlist

        Returns a list of items this Playlist iterates over, including
        sub-Playlists and paths to media files.
        """
        return list(self.__children)

    @property
    def options(self):
        """Values that modify the Playlist's iteration rules

        Returns a dictioinary of the options in use by the Playlist.
        """
        return dict(self.__options)

    @staticmethod
    def __resolve_path(path):
        root = Path(get_config("media_root"))
        resolved_path = (root / path).resolve()
        if resolved_path.exists():
            resolved_paths = [resolved_path]
        else:
            resolved_paths = sorted(p for p in root.glob(path))
            resolved_paths = [p for p in resolved_paths if p.is_file()]
            if not resolved_paths:
                warnings.warn(
                    f'file "{resolved_path}" not found and has no matches as a glob pattern'
                )

        if get_config("media_exts"):
            supported_paths = []
            for resolved_path in resolved_paths:
                if resolved_path.suffix.lower().lstrip(".") in get_config("media_exts"):
                    supported_paths.append(resolved_path)
                else:
                    warnings.warn(
                        f'file "{resolved_path}" has unsupported extension and will be skipped'
                    )
        else:
            supported_paths = resolved_paths
        return supported_paths

    def __create_sequence(self):
        processed_children = []
        for child in self.children:
            if isinstance(child, Playlist):
                # pylint: disable=protected-access
                # Reason: accessing a protected member of another
                # instance of the same class
                processed_children.append(child.__create_sequence())
            if isinstance(child, str):
                processed_children.extend(self.__resolve_path(child))

        if self.options["loop"]:
            if self.options["shuffle"]:
                return seq.ShuffledLoopingSequence(
                    *processed_children,
                    weight=self.options["weight"],
                    recent_portion=self.options["recent_portion"],
                )
            return seq.OrderedLoopingSequence(
                *processed_children,
                weight=self.options["weight"],
                portion=self.options["portion"],
                count=self.options["count"],
            )

        return seq.FiniteSequence(
            *processed_children,
            weight=self.options["weight"],
            portion=self.options["portion"],
            count=self.options["count"],
            shuffle=self.options["shuffle"],
        )

    def __iter__(self):
        return self.PlaylistState(iter(self.__create_sequence()))

    def save_xml(self, filepath=None):
        """Save the Playlist to an xml file

        The Playlist can later be recreated using Playlist.load_xml

        Arguments
        ---------
        filepath: str or None
            The desired path of the xml file. If None, uses the value of
            the "playlist_path" config.
        """
        if filepath is None:
            filepath = get_config("playlist_path")

        root = etree.Element("nextsong")

        meta = etree.Element("meta")
        root.append(meta)

        def to_attributes(options):
            attributes = {}
            for key, val in options.items():
                if val is None:
                    continue
                if isinstance(val, bool):
                    if val is True:
                        attributes[key] = "true"
                        continue
                    if val is False:
                        attributes[key] = "false"
                        continue
                if isinstance(val, (int, float)):
                    attributes[key] = str(val)
                    continue
                if isinstance(val, (tuple, list)):
                    attributes[key] = " ".join(str(x) for x in val)
                    continue
                warnings.warn(f'could not serialize option "{key}" with value "{val}"')
            return attributes

        def to_elem(node):
            if isinstance(node, str):
                elem = etree.Element("path")
                elem.text = node
                return elem
            if isinstance(node, Playlist):
                elem = etree.Element("playlist", **to_attributes(node.options))
                for child in node.children:
                    subelem = to_elem(child)
                    elem.append(subelem)
                return elem
            raise ValueError(f"Unexpected Playlist item: {node}")

        elem = to_elem(self)
        root.append(elem)

        tree = etree.ElementTree(root)
        tree.write(filepath, pretty_print=True)

    @staticmethod
    def load_xml(filepath=None):
        """Create a Playlist instance loaded from an xml file

        The xml file should follow the format of one created by
        Playlist.save_xml.

        Arguments
        ---------
        filepath: str or None
            The path to the xml file. If None, uses the value of the
            "playlist_path" config.
        """
        if filepath is None:
            filepath = get_config("playlist_path")

        def to_options(attributes):
            options = {}
            for key, val in attributes.items():
                if val == "true":
                    options[key] = True
                    continue
                if val == "false":
                    options[key] = False
                    continue
                tokens = val.split(" ")
                parsed_tokens = []
                for token in tokens:
                    parse_type = float if "." in token else int
                    try:
                        parsed_tokens.append(parse_type(token))
                    except ValueError:
                        warnings.warn(
                            f'could not deserialize attribute "{key}" with value "{val}"'
                        )
                        continue
                if len(parsed_tokens) == 1:
                    options[key] = parsed_tokens[0]
                else:
                    options[key] = parsed_tokens
            return options

        def to_node(elem):
            if elem.tag.lower() == "path":
                return elem.text
            if elem.tag.lower() == "playlist":
                children = [to_node(x) for x in elem]
                children = [x for x in children if x is not None]
                options = to_options(elem.attrib)
                return Playlist(*children, **options)
            warnings.warn(f'unexpected tag "{elem.tag}"')
            return None

        try:
            tree = etree.parse(filepath)
        except OSError:
            # lxml.etree only raises a basic OSError. Try opening
            # ourselves to trigger a more detailed error.
            with open(filepath, "r"):
                pass
        root = tree.getroot()
        if root.tag.lower() == "playlist":
            elem = root
        else:
            subelems = [x for x in root if x.tag.lower() == "playlist"]
            if len(subelems) != 1:
                raise ValueError("could not find element with playlist tag")
            elem = subelems[0]

        return to_node(elem)

    load_state = PlaylistState.load


def ensure_state(*, state_path=None, playlist_path=None, new_state=None):
    """Loads a Playlist.PlaylistState or creates a new one

    This function will either load an existing PlaylistState from the
    filesystem, or if it doesn't exist, load a Playlist from the
    filesystem and use it to create a new PlaylistState.

    Arguments
    ---------
    state_path: str or None
        The path to the PlaylistState pickle file. If None, uses the
        value of the "state_path" config.
    playlist_path: str or None
        The path to the playlist xml file. This is used to create the
        PlaylistState if the pickle file doesn't already exist. If None,
        uses the value of the "playlist_path" config.
    new_state: bool or None
        If True, forces the creation of a new PlaylistState from a
        Playlist. If None, uses the value of the "new_state" config.
    """
    if new_state is None:
        new_state = get_config("new_state")

    state = None

    if not new_state:
        try:
            return Playlist.load_state(state_path)
        except FileNotFoundError:
            pass

    if state is None:
        playlist = Playlist.load_xml(playlist_path)
        state = iter(playlist)
        state.save(state_path)

    return state



"""Implementation of config subpackage"""

__all__ = ["get", "Config"]

import os
from collections import defaultdict


def _parse_exts(exts):
    if isinstance(exts, str):
        exts = exts.split()
    return [ext.lower().lstrip(".") for ext in exts]


config_types = defaultdict(
    lambda: str,
    **{
        "new_state": bool,
        "media_exts": _parse_exts,
    },
)


def _parse_config(config):
    return {k: config_types[k](v) for k, v in config.items() if v is not None}


default_config = {
    "priority": 0,
    "values": {
        "media_root": "./media",
        "media_exts": [],
        "playlist_path": "./nextsong.xml",
        "state_path": "./state.pickle",
        "new_state": False,
    },
}

env_config = {
    "priority": 10,
    "values": _parse_config(
        {
            "media_root": os.getenv("NEXTSONG_MEDIA_ROOT", None),
            "media_exts": os.getenv("NEXTSONG_MEDIA_EXTS", None),
            "playlist_path": os.getenv("NEXTSONG_PLAYLIST_PATH", None),
            "state_path": os.getenv("NEXTSONG_STATE_PATH", None),
            "new_state": os.getenv("NEXTSONG_NEW_STATE", None),
        }
    ),
}

config_stack = [default_config, env_config]


def get(key):
    """Get a config value by name"""
    items = sorted((x["priority"], i, x["values"]) for i, x in enumerate(config_stack))
    for _, _, values in reversed(items):
        if key in values:
            return values[key]
    raise KeyError(f'config "{key}" is not set')


class Config:
    """Overrides the global configuration

    This class is a context manager that overrides some or all of the
    global configuration inside of a 'with' block

    """

    def __init__(self, priority=20, **values):
        """Create a Config instance

        The Config instance should be used in a 'with' statement and
        is only active for the duration of the 'with' block. The
        priority argument can be used to tune the overriding of other
        configs. See the module-level docstring for details. Other
        keyword arguments are specific config values.

        """
        self.__config = {
            "priority": priority,
            "values": _parse_config(values),
        }

    def __enter__(self):
        config_stack.append(self.__config)

    def __exit__(self, exc_type, exc_value, traceback):
        config_stack.pop()



"""This module provides a global configuration to the package

Configuration is sourced from defaults, environment variables, and user
overrides using this module's Config class. A config value can be
queried from anywhere using this module's 'get' function.

Config values
-------------
media_root:
    Used as the root when resolving relative paths to media files in a
    Playlist. Defaults to "./media".
media_exts:
    Lists permitted file extensions for media files. If none are given
    the related checks are skipped. Defaults to None.
playlist_path:
    The default filepath used when saving/loading a playlist with the
    Playlist.load_xml and Playlist.save_xml methods. Defaults to
    "./nextsong.xml".
state_path:
    The default filepath used when pickling/unpickling a playlist's
    state. Defaults to "./state.pickle".
new_state:
    If True, an existing pickle file is ignored and playlist state
    should be recreated from scratch from the playlist. Defaults to
    False.

Environment variables
---------------------
NEXTSONG_MEDIA_ROOT:
    Overrides media_root config
NEXTSONG_MEDIA_EXTS:
    Space-separated list; overrides media_exts config
NEXTSONG_PLAYLIST_PATH:
    Overrides playlist_path config
NEXTSONG_STATE_PATH:
    Overrides state_path config
NEXTSONG_NEW_STATE:
    False if empty string, True if anything else; overrides new_state

Details
-------
A configuration value is looked up by searching through a global stack
of config dictionaries for an item with matching key. New config
dictionaries can be pushed to the stack to override existing ones. The
stack is initialized with a dictionary containing default values for
all known config keys, followed by a dictionary containing overrides
from any known environment variables. Additional configs can be pushed
to the stack using this module's Config class to initiate a 'with'
block. The config is pushed and popped at the beginning and end of the
'with' block, respectively.

Configurations also include a priority determining who overrides who. A
config with larger number priority overrides a config with smaller
number priority, and the position on the stack is the tiebreaker. For
example, this can be used to create a user config that overrides the
defaults, but not environment variables by using a priority of 5
(between the default priority of 0 and environment variable priority of
10).

"""

from ._config import *



"""Functions implementing command line executables"""
import argparse

from nextsong.config import get as get_cfg
from nextsong import ensure_state
import nextsong as nextsong_pkg


def nextsong():
    """Gets the next song in the playlist

    Run with --help for more info
    """

    parser = argparse.ArgumentParser(prog="nextsong")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {nextsong_pkg.__version__}",
    )
    parser.add_argument(
        "-m",
        "--media-root",
        help="root directory for media files [%(default)s]",
        default=get_cfg("media_root"),
    )
    parser.add_argument(
        "-e",
        "--media-ext",
        action="append",
        help="permit file extension, repeatable [%(default)s]",
        default=get_cfg("media_exts"),
    )
    parser.add_argument(
        "-p",
        "--playlist",
        help="xml playlist filepath [%(default)s]",
        default=get_cfg("playlist_path"),
    )
    parser.add_argument(
        "-s",
        "--state",
        help="playlist state filepath [%(default)s]",
        default=get_cfg("state_path"),
    )
    parser.add_argument(
        "-n",
        "--new-state",
        action="store_true",
        help="start playlist over, ignoring existing state file [%(default)s]",
        default=get_cfg("new_state"),
    )
    args = parser.parse_args()

    with nextsong_pkg.config.Config(
        media_root=args.media_root,
        media_exts=args.media_ext or None,
        playlist_path=args.playlist,
        state_path=args.state,
        new_state=args.new_state,
    ):
        with ensure_state() as state:
            try:
                print(next(state))
            except StopIteration:
                print()
